#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "access/heapam.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"
#include "utils/regexp.h"

PG_MODULE_MAGIC;

Datum regexp_matches(PG_FUNCTION_ARGS);
Datum regexp_split(PG_FUNCTION_ARGS);


typedef struct re_comp_flags {
	bool		  return_pre_and_post;
	bool		  glob;
	int			  cflags;
} re_comp_flags;

static void parse_re_comp_flags(re_comp_flags *flags, text *o);

typedef struct regexp_matches_ctx {
	text		* orig_str;
	size_t		  orig_len;
	pg_wchar	* wide_str;
	size_t		  wide_len;
	regex_t		* cpattern;
	size_t		  offset;
	regmatch_t	* pmatch;

	/* flag options */
	re_comp_flags flags;

	/* return type info */
	TupleDesc	  rettupdesc;
	Oid			  rettype;
	TypeFuncClass typefunc;

	/* text type info */
	Oid		  param_type;
	int16		  typlen;
	bool		  typbyval;
	char		  typalign;
} regexp_matches_ctx;



#define RETURN_DATUM_MAYBE_SRF(_result) \
	do { \
		if (doing_srf) \
			SRF_RETURN_NEXT(funcctx, _result); \
		else \
			PG_RETURN_DATUM(_result); \
	} while (0)

#define RETURN_DONE_MAYBE_SRF() \
	do { \
		if (doing_srf) \
			SRF_RETURN_DONE(funcctx); \
		else \
			PG_RETURN_NULL(); \
	} while (0)

#define GETARG_TEXT_P_COPY_IF_SRF(_argno) \
	(doing_srf ? PG_GETARG_TEXT_P_COPY(_argno) : PG_GETARG_TEXT_P(_argno))

PG_FUNCTION_INFO_V1(regexp_matches);
Datum
regexp_matches(PG_FUNCTION_ARGS)
{
	FuncCallContext		*funcctx = NULL;
	MemoryContext		 oldcontext = NULL;
	regexp_matches_ctx	*matchctx = NULL;
	re_comp_flags		 flags;
	bool doing_srf = (fcinfo->resultinfo != NULL && IsA(fcinfo->resultinfo, ReturnSetInfo));

	if (!doing_srf || SRF_IS_FIRSTCALL())
	{
		text *p;
		text *o = (PG_NARGS() > 2 ? PG_GETARG_TEXT_P(2) : NULL);

		parse_re_comp_flags(&flags, o);

		if (!flags.glob)
			doing_srf = false;

		if (doing_srf)
		{
			funcctx = SRF_FIRSTCALL_INIT();
			oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		}

		matchctx = palloc0 (sizeof(regexp_matches_ctx));

		/* this needs to be visible so copy it if SRF */
		matchctx->orig_str = GETARG_TEXT_P_COPY_IF_SRF(0);
		matchctx->orig_len = VARSIZE(matchctx->orig_str) - VARHDRSZ;

		p = PG_GETARG_TEXT_P(1);

		memcpy(&matchctx->flags, &flags, sizeof(re_comp_flags));

		matchctx->cpattern = RE_compile_and_cache(p, matchctx->flags.cflags);
		matchctx->pmatch = palloc0(sizeof(regmatch_t) * (matchctx->cpattern->re_nsub + 1));
		matchctx->offset = 0;

		/* get text type oid, too lazy to do it some other way */
		matchctx->param_type = get_fn_expr_argtype(fcinfo->flinfo, 0);
		get_typlenbyvalalign(matchctx->param_type, &matchctx->typlen, &matchctx->typbyval, &matchctx->typalign);

		matchctx->typefunc = get_call_result_type(fcinfo, &matchctx->rettype, &matchctx->rettupdesc);

		if (matchctx->typefunc == TYPEFUNC_COMPOSITE)
			matchctx->rettupdesc = BlessTupleDesc(matchctx->rettupdesc);

		matchctx->wide_str = (pg_wchar *) palloc((matchctx->orig_len + 1) * sizeof(pg_wchar));
		matchctx->wide_len = pg_mb2wchar_with_len(VARDATA(matchctx->orig_str), matchctx->wide_str, matchctx->orig_len);

		matchctx->pmatch[0].rm_so = -1;
		/* both < 0 but not equal */
		matchctx->pmatch[0].rm_eo = -2;

		if (doing_srf)
		{
			MemoryContextSwitchTo(oldcontext);
			funcctx->user_fctx = (void*)matchctx;

			/* Avoid run-away function by making sure we never iterate more than the length of the text */
			funcctx->max_calls = (matchctx->flags.glob ? (3 * matchctx->orig_len) : 1);
		}
	}

	if (doing_srf)
	{
		funcctx = SRF_PERCALL_SETUP();
		matchctx = (regexp_matches_ctx*)(funcctx->user_fctx);

		if (funcctx->call_cntr >= funcctx->max_calls)
		{
			if (funcctx->max_calls != 1)
				ereport(ERROR,
						(errcode(ERRCODE_INTERNAL_ERROR),
						 errmsg("set returning match function terminated after iterating %d times", funcctx->call_cntr)));
			SRF_RETURN_DONE(funcctx);
		}
	}

	if (matchctx->offset < matchctx->orig_len)
	{
		if (matchctx->pmatch[0].rm_so == matchctx->pmatch[0].rm_eo)
			matchctx->offset++;

		if (RE_wchar_execute(matchctx->cpattern,
					matchctx->wide_str,
					matchctx->wide_len,
					matchctx->offset,
					matchctx->cpattern->re_nsub + 1,
					matchctx->pmatch))
		{
			ArrayType * resultarray;
			Datum result;
			Datum elems[matchctx->cpattern->re_nsub];
			bool nulls[matchctx->cpattern->re_nsub];
			size_t i;
			int ndims = 1;
			int dims[1] = {matchctx->cpattern->re_nsub};
			int lbs[1] = {1};

			for (i = 0; i < matchctx->cpattern->re_nsub; ++i)
			{
				int so = matchctx->pmatch[i+1].rm_so,
					eo = matchctx->pmatch[i+1].rm_eo;

				if (so < 0 || eo < 0)
				{
					elems[i] = 0;
					nulls[i] = true;
				}
				else
				{
					elems[i] = DirectFunctionCall3(text_substr,
							PointerGetDatum(matchctx->orig_str),
							Int32GetDatum(so + 1),
							Int32GetDatum(eo - so));
					nulls[i] = false;
				}
			}

			resultarray = construct_md_array(elems, nulls, ndims, dims, lbs,
					matchctx->param_type, matchctx->typlen, matchctx->typbyval, matchctx->typalign);

			if (matchctx->typefunc == TYPEFUNC_COMPOSITE)
			{
				Datum vals[4];
				bool  valnulls[4] = {true, true, true, true};

				if (matchctx->flags.return_pre_and_post)
				{
					valnulls[0] = valnulls[3] = false;

					vals[0] = DirectFunctionCall3(text_substr,
							PointerGetDatum(matchctx->orig_str),
							Int32GetDatum(1),
							Int32GetDatum(matchctx->pmatch[0].rm_so));

					vals[3] = DirectFunctionCall2(text_substr_no_len,
							PointerGetDatum(matchctx->orig_str),
							Int32GetDatum(matchctx->pmatch[0].rm_eo + 1));
				}

				vals[1] = DirectFunctionCall3(text_substr,
						PointerGetDatum(matchctx->orig_str),
						Int32GetDatum(matchctx->pmatch[0].rm_so + 1),
						Int32GetDatum(matchctx->pmatch[0].rm_eo - matchctx->pmatch[0].rm_so));
				valnulls[1] = false;


				vals[2] = PointerGetDatum(resultarray);
				valnulls[2] = false;

				result = HeapTupleGetDatum(
						heap_form_tuple (matchctx->rettupdesc, vals, valnulls));
			}
			else
			{
				result = PointerGetDatum(resultarray);
			}

			matchctx->offset = matchctx->pmatch[0].rm_eo;
			RETURN_DATUM_MAYBE_SRF(result);
		}
		/* else fall through and return done (or null) */
	}

	RETURN_DONE_MAYBE_SRF();
}


typedef struct regexp_split_ctx {
	text		* orig_str;
	size_t		  orig_len;
	pg_wchar	* wide_str;
	size_t		  wide_len;
	regex_t		* cpattern;
	regmatch_t	  match;
	size_t		  offset;

	/* flag options */
	re_comp_flags flags;
} regexp_split_ctx;

PG_FUNCTION_INFO_V1(regexp_split);
Datum
regexp_split(PG_FUNCTION_ARGS)
{
	FuncCallContext  *funcctx;
	MemoryContext     oldcontext;
	regexp_split_ctx *splitctx;

	if (SRF_IS_FIRSTCALL())
	{
		text * o = (PG_NARGS() > 2 ? PG_GETARG_TEXT_P(2) : NULL);

		funcctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		splitctx = (regexp_split_ctx *) palloc(sizeof(regexp_split_ctx));

		splitctx->orig_str = PG_GETARG_TEXT_P_COPY(0);
		splitctx->orig_len = VARSIZE(splitctx->orig_str) - VARHDRSZ;

		parse_re_comp_flags(&splitctx->flags, o);
		if (splitctx->flags.glob)
		{
			ereport(ERROR,
					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
					 errmsg("regexp_split does not support the global option")));
		}

		if (splitctx->flags.return_pre_and_post)
		{
			ereport(ERROR,
					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
					 errmsg("regexp_split does not support the return pre and postmatch option")));
		}


		splitctx->cpattern = RE_compile_and_cache(PG_GETARG_TEXT_P(1), splitctx->flags.cflags);

		splitctx->wide_str = (pg_wchar *) palloc((splitctx->orig_len + 1) * sizeof(pg_wchar));
		splitctx->wide_len = pg_mb2wchar_with_len(VARDATA(splitctx->orig_str), splitctx->wide_str, splitctx->orig_len);

		splitctx->offset = 0;

		splitctx->match.rm_so = -1;
		/* both < 0 but not equal */
		splitctx->match.rm_eo = -2;

		MemoryContextSwitchTo(oldcontext);
		funcctx->user_fctx = (void*)splitctx;

		/* Avoid run-away function by making sure we never iterate more than the length of the text */
		funcctx->max_calls = 3 * splitctx->orig_len;
	}

	funcctx = SRF_PERCALL_SETUP();
	splitctx = (regexp_split_ctx*)(funcctx->user_fctx);

	if (funcctx->call_cntr > funcctx->max_calls)
	{
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("set returning split function terminated after iterating %d times", funcctx->call_cntr)));
		SRF_RETURN_DONE(funcctx);
	}

	if (splitctx->offset < splitctx->orig_len)
	{
		regmatch_t * pmatch = &(splitctx->match);
		bool first_match = (pmatch->rm_so < 0 || pmatch->rm_eo < 0);
		bool incremented_offset = false;
		do
		{
			if (pmatch->rm_so == pmatch->rm_eo)
			{
				splitctx->offset++;
				incremented_offset = true;
			}

			if (RE_wchar_execute(splitctx->cpattern,
						splitctx->wide_str,
						splitctx->wide_len,
						splitctx->offset,
						1,
						pmatch))
			{
				if (!((first_match && pmatch->rm_so == pmatch->rm_eo) || (pmatch->rm_so == splitctx->offset && !incremented_offset)))
				{
					Datum result = DirectFunctionCall3(text_substr,
							PointerGetDatum(splitctx->orig_str),
							Int32GetDatum(splitctx->offset + (incremented_offset ? 0 : 1)),
							Int32GetDatum(pmatch->rm_so - splitctx->offset + (incremented_offset ? 1 : 0)));
					splitctx->offset = pmatch->rm_eo;
					SRF_RETURN_NEXT(funcctx, result);
				}
			}
			else
			{
				/* no more matches, return rest of string */
				Datum result = DirectFunctionCall2(text_substr_no_len,
						PointerGetDatum(splitctx->orig_str),
						Int32GetDatum(splitctx->offset + (incremented_offset ? 0 : 1)));
				/* so we know we're done next time thru */
				splitctx->offset = splitctx->orig_len;
				SRF_RETURN_NEXT(funcctx, result);
			}
			first_match = false;
		} while (1);
	}

	SRF_RETURN_DONE(funcctx);
}

static void parse_re_comp_flags(re_comp_flags *flags, text *o)
{
	memset(flags, 0, sizeof(re_comp_flags));
	flags->cflags = regex_flavor;

	if (o)
	{
		char * opt_p = VARDATA(o);
		size_t opt_len = (VARSIZE(o) - VARHDRSZ);
		size_t i;

		for (i = 0; i < opt_len; i++)
		{
			switch (opt_p[i])
			{
				case 'g':
					flags->glob = true;
					break;
				case 'i':
					flags->cflags |= REG_ICASE;
					break;
				case 'm':
				case 'n':
					flags->cflags |= REG_NEWLINE;
					break;
				case 'p':
					flags->cflags |= REG_NLSTOP;
					flags->cflags &= ~REG_NLANCH;
					break;
				case 'r':
					flags->return_pre_and_post = true;
					break;
				case 'w':
					flags->cflags &= ~REG_NLSTOP;
					flags->cflags |= REG_NLANCH;
					break;
				case 'x':
					flags->cflags |= REG_EXPANDED;
					break;
				default:
					ereport(ERROR,
							(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
							 errmsg("invalid regexp option: %c", opt_p[i])));
					break;
			}
		}
	}
}
